@@ -26,10 +26,11 @@ def markdown(text):
 LIST_SPECIFIERS = ('*', '+', '-')
 ORDERED_LIST = 1
 UNORDERED_LIST = 2
-IS_LIST = ORDERED_LIST & UNORDERED_LIST
 
 # regexs
 # FIXME I hate using all these negative lookbehinds!
+# Actually, I don't like using re at all... but it's comparatively faster.
+
 LIST_1 = re.compile('{0}{0}(.+?{0}?){0}{0}'.format('(?<!\\\)\*'))
 LIST_2 = re.compile('{0}(.+?){0}'.format('(?<!\\\)\*'))
 LIST_3 = re.compile('{0}{0}(.+?{0}?){0}{0}'.format('(?<!\\\)\_'))
@@ -43,17 +44,29 @@ AMP = re.compile('&(?!\w+;)')
 HR = re.compile('^([- ]{3,}|[* ]{3,}|[_ ]{3,})$')
 HREF = re.compile('<(http\S+)>')
 HREF_inline = lambda m:'<a href="{0}">{0}</a>'.format(m.group(1))
-URL = re.compile('\[(.+?)\]\((.*?)(?:\s+\"(.*?)\")?\s*\)', re.DOTALL)
-URL_a = lambda m:'<a href="{1}"{2}>{0}</a>'.format(m.group(1), m.group(2),
-            ' title="{0}"'.format(escape(m.group(3))) if m.group(3) else '')
-URL_REF = re.compile('(?<!\\\)\[([^\[\]\\\]+?)\](?:\s*\[([^\[\]\\\]*?)\])?')
+
+# [text](/url/ "title")
+URL = re.compile('((?<!\\\)\!)?\[(.+?)\]\((.*?)(?:\s+\"(.*?)\")?\s*\)',
+        re.DOTALL)
+def URL_a (m):
+    if m.group(1): return IMG_img(m)
+    return '<a href="{1}"{2}>{0}</a>'.format(m.group(2), m.group(3),
+            ' title="{0}"'.format(escape(m.group(4))) if m.group(4) else '')
+
+# [id][ID]
+URL_REF = re.compile('((?<!\\\)\!)?(?<!\\\)\[([^\[\]\\\]+?|(?:.*?\[.*?\].*?)+)\](?:\s*\[([^\[\]\\\]*?)\])?', re.DOTALL)
+
+# [id]: /url/ "title"
 URL_DEF = re.compile('\s{0,3}\[([^\[\]\\\]+?)\]:\s*([^ \"\\\]+)\s*(?:\"([^\[\]\\\]+)\")?')
 
-IMG = re.compile('(?<!\\\)\!\[(.+?)\]\((.*?)(?:\s+\"(.*?)\")?\s*\)', re.DOTALL)
-IMG_img = lambda m:'<img src="{1}"{2} alt="{0}" />'.format(m.group(1),
-        m.group(2),
-        ' title="{0}"'.format(escape(m.group(3))) if m.group(3) else '')
-IMG_REF = re.compile('(?<!\\\)\!(?<!\\\)\[([^\[\]\\\]+?)\](?:\s*\[([^\[\]\\\]*?)\])?')
+# ![alt](/url/ "title")
+IMG = re.compile('((?<!\\\)\!)\[(.+?)\]\((.*?)(?:\s+\"(.*?)\")?\s*\)', re.DOTALL)
+IMG_img = lambda m:'<img src="{1}"{2} alt="{0}" />'.format(m.group(2),
+        m.group(3),
+        ' title="{0}"'.format(escape(m.group(4))) if m.group(4) else '')
+
+# [id]: /url/to/image/ "title"
+IMG_REF = re.compile('((?<!\\\)\!)(?<!\\\)\[([^\[\]\\\]+?)\](?:\s*\[([^\[\]\\\]*?)\])?')
 
 FIND_CODE_SPAN = re.compile(r'(?<!\\)`')
 FIND_INLINE_HTML = re.compile(r'(?<!\\)<')
@@ -124,7 +137,7 @@ class Markdown:
                 # block. (indention level less than 4)
                 if i < 4:
                     # atx-style header
-                    if b and (not stripped.strip('-') or\
+                    if b and not last_is_block and (not stripped.strip('-') or\
                               not stripped.strip('=')):
                         ht = b.pop()
                         if '\n' in ht:
@@ -191,9 +204,10 @@ class Markdown:
                             last_is_block = True
                             block_type = BLOCKQUOTE
 
-                    elif last_was_blank:
+                    elif last_was_blank and not (last_is_block and i > 1):
                         # No block defs were made and the last line was blank.
                         # That means this line starts a new paragraph.
+                        # UNLESS it's part of a list item.
                         if b:
                             yield Node(b, self, block_indent)
                             b = []
@@ -336,7 +350,6 @@ class Node:
         return True
 
 
-
     def markdown_lines(self, strio, lines=None, wrap=None):
         if not lines: lines = self.lines
         first = True
@@ -397,30 +410,48 @@ class Node:
 
                 if '[' in l:
                     if '!' in l:
-                        l = IMG.sub(IMG_img, l)
+                        _has_image = True
                         def IMG_REF_a(m):
-                            v, id = m.groups()
+                            _, v, id = m.groups()
                             id = ' '.join([s.strip() for s in (id or v).split('\n')])
                             if id not in self.markdown.urls:
                                 return m.group(0)
                             url, title = self.markdown.urls[id]
                             return '<img src="{1}"{2} alt="{0}" \>'.format(v,
-                             url, ' title="{0}"'.format(title) if title else '')
-                        l = IMG_REF.sub(IMG_REF_a, l)
+                            url, ' title="{0}"'.format(title) if title else '')
+                    else:
+                        _has_image = False
 
 
-                    l = URL.sub(URL_a, l)
-
                     def URL_REF_a(m):
-                        v, id = m.groups()
-                        id = ' '.join([s.strip() for s in (id or v).split('\n')])
+                        im, v, id = m.groups()
+                        if im:
+                            return IMG_REF_a(m)
+                        id = ' '.join(
+                            [s.strip() for s in (id or v).split('\n')])
+                        _replaced = False
+                        if '[' in v and ']' in v:
+                            if _has_image:
+                                v = IMG_REF.sub(IMG_REF_a, v)
+                            v = URL_REF.sub(URL_REF_a, v)
+                            _replaced = True
                         if id not in self.markdown.urls:
-                            return m.group(0)
+                            if _replaced:
+                                return '['+v+']'
+                            else:
+                                return m.group(0)
                         url, title = self.markdown.urls[id]
                         return '<a href="{1}"{2}>{0}</a>'.format(v, url,
                             ' title="{0}"'.format(title) if title else '')
+
+                    # The order of these subs is important.
+                    l = URL.sub(URL_a, l)
                     l = URL_REF.sub(URL_REF_a, l)
 
+                    if _has_image:
+                        l = IMG.sub(IMG_img, l)
+                        l = IMG_REF.sub(IMG_REF_a, l)
+
                 if has_bs:
                     l = FIND_BS.sub('', l)
                 strio.write(l)
@@ -438,7 +469,5 @@ class Node:
                     if not r:
                         strio.write(l.replace('<', '&lt;'))
             
-
-
         if wrap:
             strio.write('</'+wrap+'>\n')
